#!/usr/bin/python
# -*- coding: utf-8 -*-

"""Check if it is possible to install a specified module."""

from __future__ import unicode_literals, print_function

import argparse
import itertools
import json
import os
import subprocess
import sys


KEYS = {'installer': 'CHECKINSTALL_INSTALLER',
        'package': 'CHECKINSTALL_PACKAGE',
        'version': 'CHECKINSTALL_PACKAGE_VERSION',
        'dependencies': 'CHECKINSTALL_DEPS',
        'separator': 'CHECKINSTALL_DEPS_SEPARATOR'}
"""Configuration keys, usable as names of environment variables."""

DEFAULT = {KEYS['installer']: 'pip',
           KEYS['dependencies']: '',
           KEYS['separator']: None}
"""Default attribute values."""

EXIT_CODES = {'dependencies': 2,
              'package': 1,
              'required_args': 0xFF}
"""Exit codes indicating the type of failure."""


class DependencyInstallFailure(OSError):
    """Exception raised when the dependecy installation fails for any reason."""
    __slots__ = ['cmd', 'log']

    def __init__(self, exitcode, cmd, log = ''):
        super(self.__class__, self).__init__(exitcode, log, cmd)

        self.cmd = cmd
        self.log = log


class PackageInstallFailure(OSError):
    """Exception raised when the module installation fails for any reason."""
    __slots__ = ['package', 'log']

    def __init__(self, exitcode, package, log=''):
        super(self.__class__, self).__init__(exitcode, log, package)

        self.package = package
        self.log = log


def install_dependencies(dependencies):
    """Attempts to install provided module dependencies.

    Keyword arguments:
        dependencies -- list of dependencies

    Returns: None
    Throws: DependencyInstallFailure.
    """

    if len(dependencies) == 0: return

    # Yum or DNF?
    command = ['dnf' if os.path.isfile('/usr/bin/dnf') else 'yum',
               '--assumeyes', 'install']

    process = subprocess.Popen(command + dependencies,
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    output, _ = process.communicate()

    if process.returncode != 0:
        raise DependencyInstallFailure(
                exitcode=process.returncode,
                cmd=' '.join(command + dependencies),
                log=output)


def pip(package, version = None):
    """Install specified package using pip.

    Returns: The pip output (log).
    Throws: PackageInstallFailure.
    """

    command = ['pip', 'install',
              '=='.join(itertools.ifilter(None, (package, version)))]

    # Extra options
    command += '--exists-action i'.split()  # Ignore existing packages

    process = subprocess.Popen(command, stdout=subprocess.PIPE)
    output, _ = process.communicate()

    if process.returncode != 0:
        raise PackageInstallFailure(
                exitcode=process.returncode,
                package=package,
                log=output)
    else:
        return output


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(description='Automated installation validator.')

    argparser.add_argument('-d', '--deps', metavar='package-version', nargs='+',
                           dest=KEYS['dependencies'], default=argparse.SUPPRESS,
                           help='Predetermined module dependencies')
    argparser.add_argument('-s', '--deps-separator', metavar='SEP',
                           dest=KEYS['separator'],
                           help='Separator of the dependecies (whitespace by default)')

    argparser.add_argument(KEYS['installer'], metavar='installer',
                           nargs='?', default=argparse.SUPPRESS,
                           help='Module installer')
    argparser.add_argument(KEYS['package'], metavar='package',
                           nargs='?', default=argparse.SUPPRESS,
                           help='Module name')
    argparser.add_argument(KEYS['version'], metavar='version',
                           nargs='?', default=argparse.SUPPRESS,
                           help='Module version')

    # No collections.ChainMap in Python2 :(
    options = DEFAULT.copy()
    options.update({k:os.environ[k] for k in KEYS.itervalues() if k in os.environ})
    options.update(vars(argparser.parse_args()))

    output = {'success': False}

    try:
        dependencies, separator = (options[KEYS['dependencies']],
                                   options[KEYS['separator']])
        install_dependencies(dependencies.split(separator))

        installer, package, version = (options[KEYS['installer']],
                                       options[KEYS['package']],
                                       options[KEYS['version']]
                                            if KEYS['version'] in options
                                            else None)
        if installer == 'pip':
            output['log'] = pip(package, version)
        else:
            raise NotImplemented('Only pip so far')

        output['success'] = True

    except KeyError as exc:
        key = unicode(exc).lstrip("u'").rstrip("'")  # Where the hell I got the "u''" as part of the string !?
        if key.startswith('CHECKINSTALL'):
            print('Missing required argument: '+key, file=sys.stderr)
            raise SystemExit(EXIT_CODES['required_args'])

    except DependencyInstallFailure as failure:
        output['log'] = failure.log
        raise SystemExit(EXIT_CODES['dependencies'])

    except PackageInstallFailure as failure:
        output['log'] = failure.log
        raise SystemExit(EXIT_CODES['package'])

    finally:
        json.dump(output, sys.stdout, ensure_ascii=False)
